/*  Audacious-DUMB Plug-in to enable audacious to use DUMB for IT/XM/S3M/MOD files
 *  Copyright (C) 2006 Christian Birchinger
 *  Audacious port based on previous plugins for XMMS and BMP
 *  Copyright (C) 2002-2003  Ben Davis
 *  Incorporates code from the wav plug-in,
 *  Copyright (C) 1998-2000  Peter Alm, Mikael Alm, Olle Hallnas, Thomas Nilsson and 4Front Technologies
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "audacious-dumb.h"

#include <audacious/configdb.h>

#include <gtk/gtk.h>
#include "interface.h"
#include "support.h"

#if 0
typedef struct
{
        void *handle;           /* Filled in by audacious */
        char *filename;         /* Filled in by audacious */
        char *description;      /* The description that is shown in the preferences box */
        void (*init) (void);    /* Called when the plugin is loaded */
        void (*about) (void);   /* Show the about box */
        void (*configure) (void);
        int (*is_our_file) (char *filename);    /* Return 1 if the plugin can handle the file */
        GList *(*scan_dir) (char *dirname);     /* Look in Input/cdaudio/cdaudio.c to see how to use this */
        void (*play_file) (char *filename);     /* Guess what... */
        void (*stop) (void);    /* Tricky one */
        void (*pause) (short paused);   /* Pause or unpause */
        void (*seek) (int time);        /* Seek to the specified time */
        void (*set_eq) (int on, float preamp, float *bands);    /* Set the equalizer, most plugins won't be able to do this */
        int (*get_time) (void); /* Get the time, usually returns the output plugins output time */
        void (*get_volume) (int *l, int *r);    /* Input-plugin specific volume functions, just provide a NULL if */
        void (*set_volume) (int l, int r);      /*  you want the output plugin to handle it */
        void (*cleanup) (void);                 /* Called when audacious exit */
        InputVisType (*get_vis_type) (void); /* OBSOLETE, DO NOT USE! */
        void (*add_vis_pcm) (int time, AFormat fmt, int nch, int length, void *ptr); /* Send data to the visualization plugins
                                                                                        Preferably 512 samples/block */
        void (*set_info) (char *title, int length, int rate, int freq, int nch);        /* Fill in the stuff that is shown in the player window
                                                                                           set length to -1 if it's unknown. Filled in by audacious */
        void (*set_info_text) (char *text);     /* Show some text in the song title box in the main window,
                                                   call it with NULL as argument to reset it to the song title.
                                                   Filled in by audacious */
        void (*get_song_info) (char *filename, char **title, int *length);      /* Function to grab the title string */
        void (*file_info_box) (char *filename);         /* Bring up an info window for the filename passed in */
        OutputPlugin *output;   /* Handle to the current output plugin. Filled in by audacious */
}
InputPlugin;
#endif

InputPlugin duh_ip =
{
	NULL,
	NULL,
	"DUMB Plugin " VERSION,
	duh_init,
	duh_about,
	duh_configure,
	is_our_file,
	NULL,
	play_file,
	stop,
	duh_pause,
	seek,
	NULL,
	get_time,
	NULL,
	NULL,
	dumb_exit,
	NULL,
	NULL,
	NULL,
	NULL,
	get_song_info,
	file_info_box,
	NULL
};

DuhFile *duh_file = NULL;
static GThread *decode_thread;
static gboolean audio_error = FALSE;
static gboolean disable_amiga_mods = FALSE;
static gint output_frequency;
const static gint output_frequencies[]={4000, 8000, 11025, 22050, 44100, 48000, 64000, 96000};
static GtkWidget *prefs;

InputPlugin *get_iplugin_info(void)
{
	duh_ip.description = VERSIONSTRING;
	return &duh_ip;
}

static void duh_init(void)
{
	ConfigDb *db;
	db = bmp_cfg_db_open();

	bmp_cfg_db_get_bool(db, "DUMB", "disable_amiga_mods", &disable_amiga_mods);
	if (!disable_amiga_mods) /* force db creation since defined false equals none-existant */
		bmp_cfg_db_set_bool(db, "DUMB", "disable_amiga_mods", FALSE);

	bmp_cfg_db_get_int(db, "DUMB", "output_frequency", &output_frequency);	
	if (output_frequency < 4000 || output_frequency > 96000) {
		output_frequency = 44100;
		bmp_cfg_db_set_int(db, "DUMB", "output_frequency", output_frequency);
	}
		
	bmp_cfg_db_close(db);

	dumb_register_stdfiles();
	dumb_it_max_to_mix = 256;
}

static int is_our_file(char *filename)
{
	gchar *ext;

	ext = strrchr(filename, '.');
	if (ext) {
		if (!strcasecmp(ext, ".duh")) return TRUE;
		if (!strcasecmp(ext, ".it")) return TRUE;
		if (!strcasecmp(ext, ".xm")) return TRUE;
		if (!strcasecmp(ext, ".s3m")) return TRUE;
		if (!disable_amiga_mods) {
			if (!strcasecmp(ext, ".mod")) return TRUE;
		}
	}
	return FALSE;
}


static gchar *get_title(DUH *duh, gchar *filename)
{
	gchar *title;
	const gchar *mod_title;
	mod_title = duh_get_tag(duh, "TITLE");
	if (mod_title && mod_title[0])
		title = g_strdup_printf("%s - %s", g_basename(filename), mod_title);
	else
		title = g_strdup(g_basename(filename));
	return title;
}                                                                                

static void install_callbacks(DUH_SIGRENDERER *sr)
{
	DUMB_IT_SIGRENDERER *itsr = duh_get_it_sigrenderer(sr);
	dumb_it_set_loop_callback(itsr, &dumb_it_callback_terminate, NULL);
	dumb_it_set_xm_speed_zero_callback(itsr, &dumb_it_callback_terminate, NULL);
}

static void *play_loop(void *arg)
{
	char data[2048 * 2];
	int bytes, blk_size, rate;
	int actual_read;

	blk_size = 512 * (duh_file->bits_per_sample / 8) * duh_file->channels;
	rate = duh_file->samples_per_sec * duh_file->channels * (duh_file->bits_per_sample / 8);
	while (duh_file->going)
	{
		if (!duh_file->eof)
		{
			bytes = blk_size;
			if (duh_file->length - duh_file->position < bytes)
				bytes = duh_file->length - duh_file->position;
			if (bytes > 0)
			{
				actual_read = duh_render(duh_file->file, 16, 0, 1.0f, 65536.0f / duh_file->samples_per_sec, bytes / ((duh_file->bits_per_sample / 8) * duh_file->channels), data);
				actual_read *= (duh_file->bits_per_sample / 8) * duh_file->channels;
				//actual_read = fread(data, 1, bytes, duh_file->file);

				if (actual_read == 0)
				{
					// Should we do this or should we pad with silence?
					duh_file->eof = 1;
					duh_ip.output->buffer_free();
					duh_ip.output->buffer_free();
				}
				else
				{
					while ((duh_ip.output->buffer_free() < bytes) && duh_file->going)
						g_usleep(10000);

					if (!duh_file->going)
						break;
					
					produce_audio(duh_ip.output->written_time(), (duh_file->bits_per_sample == 16) ? FMT_S16_LE : FMT_U8,
					   duh_file->channels, bytes, data, &duh_file->going);

					duh_file->position += actual_read;
				}
			}
			else
			{
				duh_file->eof = TRUE;
				duh_ip.output->buffer_free();
				duh_ip.output->buffer_free();
				g_usleep(10000);
			}
		}
		else
			g_usleep(10000);
		if (duh_file->seek_to != -1)
		{
			duh_file->position = duh_file->seek_to * rate;
			duh_end_sigrenderer(duh_file->file);
			duh_file->file = duh_start_sigrenderer(duh_file->duh, 0, duh_file->channels, duh_file->seek_to << 16);
			install_callbacks(duh_file->file);
			//fseek(duh_file->file, duh_file->position + duh_file->data_offset, SEEK_SET);
			duh_ip.output->flush(duh_file->seek_to * 1000);
			duh_file->seek_to = -1;
		}

	}
	duh_end_sigrenderer(duh_file->file);
	unload_duh(duh_file->duh);
	//fclose(duh_file->file);
	g_thread_exit(NULL);
}

static void play_file(char *filename)
{
	char *name;
	int rate;

	audio_error = FALSE;

	duh_file = g_malloc(sizeof (DuhFile));
	memset(duh_file, 0, sizeof (DuhFile));

	duh_file->duh = load_duh(filename);
	if (!duh_file->duh) {
		duh_file->duh = dumb_load_it(filename);
		if (!duh_file->duh) {
			duh_file->duh = dumb_load_xm(filename);
			if (!duh_file->duh) {
				duh_file->duh = dumb_load_s3m(filename);
				if (!duh_file->duh) {
					duh_file->duh = dumb_load_mod(filename);
					if (!duh_file->duh) {
						g_free(duh_file);
						duh_file = NULL;
						return;
					}
				}
			}
		}
	}

	if ((duh_file->file = duh_start_sigrenderer(duh_file->duh, 0, 2, 0)))
	{
		install_callbacks(duh_file->file);
		duh_file->channels = 2;
		duh_file->samples_per_sec = output_frequency;
		duh_file->bits_per_sample = 16;
		duh_file->length = (long long)duh_get_length(duh_file->duh) * (output_frequency * 2 * 2) >> 16;
		duh_file->position = 0;
		duh_file->going = 1;

		if (duh_ip.output->open_audio((duh_file->bits_per_sample == 16) ? FMT_S16_LE : FMT_U8, duh_file->samples_per_sec, duh_file->channels) == 0)
		{
			audio_error = TRUE;
			duh_end_sigrenderer(duh_file->file);
			unload_duh(duh_file->duh);
			g_free(duh_file);
			duh_file = NULL;
			return;
		}
		name = get_title(duh_file->duh, filename);
		rate = duh_file->samples_per_sec * duh_file->channels * (duh_file->bits_per_sample / 8);
		duh_ip.set_info(name, 1000LL * duh_file->length / rate, 8 * rate, duh_file->samples_per_sec, duh_file->channels);
		g_free(name);
		duh_file->seek_to = -1;
		decode_thread = g_thread_create(play_loop, NULL, TRUE, NULL);
	}
}

static void stop(void)
{
	if (duh_file && duh_file->going)
	{
		duh_file->going = 0;
		g_thread_join(decode_thread);
		duh_ip.output->close_audio();
		g_free(duh_file);
		duh_file = NULL;
	}
}

static void duh_pause(short p)
{
	duh_ip.output->pause(p);
}

static void seek(int time)
{
	duh_file->seek_to = time;

	duh_file->eof = FALSE;

	while (duh_file->seek_to != -1)
		g_usleep(10000);
}

static int get_time(void)
{
	if (audio_error)
		return -2;
	if (!duh_file)
		return -1;
	if (!duh_file->going || (duh_file->eof && !duh_ip.output->buffer_playing()))
		return -1;
	else
	{
		return duh_ip.output->output_time();
	}
}

static void get_song_info(char *filename, char **title, int *length)
{
	DUH *duh = load_duh(filename);
	if (!duh) {
		duh = dumb_load_it(filename);
		if (!duh) {
			duh = dumb_load_xm(filename);
			if (!duh) {
				duh = dumb_load_s3m(filename);
				if (!duh) {
					duh = dumb_load_mod(filename);
					if (!duh) return;
				}
			}
		}
	}
	*length = (int)(duh_get_length(duh) * 1000LL >> 16);
	*title = get_title(duh, filename);
	unload_duh(duh);
}

static void file_info_box(char *filename)
{
	GtkWidget *songinfo;
	GtkStyle *style = NULL;

	GtkTextBuffer *TextBuffer;

	PangoFontDescription *font_desc = NULL;

	DUMB_IT_SIGDATA *sd;

	DUH *duh = dumb_load_it(filename);
	if (!duh) {
		duh = dumb_load_xm(filename);
		if (!duh) {
			duh = dumb_load_s3m(filename);
			if (!duh) {
				duh = dumb_load_mod(filename);
				if (!duh) return;
			}
		}
	}
	sd = duh_get_it_sigdata(duh);

	songinfo = create_songinfo ();
	if (!songinfo) return;

	font_desc = pango_font_description_from_string ("monospace 9");

	if (font_desc) {
		style = gtk_style_copy(gtk_widget_get_style(songinfo));
		style->font_desc = font_desc;
	}
	
	{
		int n = dumb_it_sd_get_n_instruments(sd);
		if (n) {
			GtkCList *l = GTK_CLIST(lookup_widget(songinfo, "songinfo_instruments_clist"));
			int i;
			if (style) gtk_widget_set_style(GTK_WIDGET(l), style);
			gtk_clist_column_titles_passive(l);
			gtk_clist_set_column_justification(l, 0, GTK_JUSTIFY_RIGHT);
			for (i = 0; i < n; i++) {
				char str[3];
				const gchar *text[] = {
					str,
					dumb_it_sd_get_instrument_name(sd, i),
					dumb_it_sd_get_instrument_filename(sd, i)
				};
				sprintf(str, "%d", i+1);
				gtk_clist_insert(l, i, (gchar **)text);
			}
		} else
			gtk_notebook_remove_page(GTK_NOTEBOOK(lookup_widget(songinfo, "songinfo_notebook")), 2);
	}

	{
		GtkCList *l = GTK_CLIST(lookup_widget(songinfo, "songinfo_samples_clist"));
		int i, n = dumb_it_sd_get_n_samples(sd);
		if (style) gtk_widget_set_style(GTK_WIDGET(l), style);
		gtk_clist_column_titles_passive(l);
		gtk_clist_set_column_justification(l, 0, GTK_JUSTIFY_RIGHT);
		for (i = 0; i < n; i++) {
			char str[3];
			const gchar *text[] = {
				str,
				dumb_it_sd_get_sample_name(sd, i),
				dumb_it_sd_get_sample_filename(sd, i)
			};
			sprintf(str, "%d", i+1);
			gtk_clist_insert(l, i, (gchar **)text);
		}
	}

	{
		GtkTextView *l = GTK_TEXT_VIEW(lookup_widget(songinfo, "songinfo_message_text"));
		const unsigned char *message = dumb_it_sd_get_song_message(sd);
		if (message) {
			unsigned char *m = strdup(message);
			int i;
			for (i = 0; m[i]; i++)
				if (m[i] == '\r') m[i] = '\n';

			TextBuffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (lookup_widget(songinfo, "songinfo_message_text")));
			if (style) gtk_widget_set_style(GTK_WIDGET(l), style);
			gtk_text_buffer_set_text(TextBuffer, m, -1);

			free(m);
		} else
			gtk_notebook_remove_page(GTK_NOTEBOOK(lookup_widget(songinfo, "songinfo_notebook")), 0);
	}

	gtk_widget_show (songinfo);
}

static void duh_configure(void) {
	if (prefs) return;
	prefs = create_prefs();
	if (!prefs) return;

	gint output_frequency_idx = 4; /* 44100 default/fallback */
	int i;
	gchar freqstring[6];

	for (i = 0; i < sizeof(output_frequencies) / sizeof(output_frequencies[0]); i++) {
		snprintf(freqstring, sizeof(freqstring), "%d", output_frequencies[i]);
		gtk_combo_box_append_text (GTK_COMBO_BOX(lookup_widget(prefs, "prefs_freq_combobox")), freqstring);
		if (output_frequencies[i] == output_frequency)
			output_frequency_idx = i;
	}
	
 	gtk_combo_box_set_active(GTK_COMBO_BOX(lookup_widget(prefs, "prefs_freq_combobox")), output_frequency_idx);

	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(lookup_widget(prefs, "prefs_checkbox_disableamiga")), disable_amiga_mods);

	gtk_signal_connect(GTK_OBJECT(prefs), "destroy",         
			   GTK_SIGNAL_FUNC(gtk_widget_destroyed), &prefs);

	gtk_signal_connect_object(GTK_OBJECT(lookup_widget(prefs, "prefs_button_cancel")), "clicked",
				  GTK_SIGNAL_FUNC(gtk_widget_destroy), 
				  GTK_OBJECT(prefs));

	gtk_signal_connect(GTK_OBJECT(lookup_widget(prefs, "prefs_button_ok")), "clicked",
				  GTK_SIGNAL_FUNC(duh_configure_ok_cb), lookup_widget(prefs, "prefs_checkbox_disableamiga"));

	gtk_widget_show(prefs);
}

static void duh_configure_ok_cb(GtkButton *w, gpointer data) {
	ConfigDb *db;

	db = bmp_cfg_db_open();

	disable_amiga_mods = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(data));
	bmp_cfg_db_set_bool(db, "DUMB", "disable_amiga_mods", disable_amiga_mods);

	output_frequency = output_frequencies[gtk_combo_box_get_active(GTK_COMBO_BOX(lookup_widget(prefs, "prefs_freq_combobox")))];
	bmp_cfg_db_set_int(db, "DUMB", "output_frequency", output_frequency);	

	bmp_cfg_db_close(db);
	gtk_widget_destroy(prefs);
}

static void duh_about(void) {
	static GtkWidget *about;

	if (about) return;

	about = xmms_show_message("About DUMB Plugin",
				  "DUMB Output Plugin " VERSION "\n"
				  "by Christian Birchinger <joker@netswarm.net>\n"
				  "\n"
				  "Based on the original XMMS plugin by Ben Davis\n"
				  "and the BMP port by Michael Doering\n"
				  "\n"
				  "Built with DUMB version " DUMB_VERSION_STR,
				  "Ok", FALSE, NULL, NULL);

	g_signal_connect(G_OBJECT(about), "destroy",
			 G_CALLBACK(gtk_widget_destroyed), &about);
}
